﻿// Copyright(c) 2015 DatumTron Corp, All Rights Reserved
// Author: Ashraf Azmi
using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Common;
using System.Data.SqlClient;
using System.Linq;
using System.Net.Sockets;
using DatumTron.Datum.Abstract;
using DatumTron.Datum.Atum;

namespace DatumTron.Apps
{
    /// <summary>
    /// Import a SQL database into a Datum Universe
    /// </summary>
    public class Importer
    {
        /// <summary>
        /// Import an SQL database into Datum Universe.
        /// </summary>
        /// <param name="atum"></param>
        /// <param name="rootName">name of the root katum of the database</param>
        /// <param name="dbs">description of the database tables</param>
        /// <param name="sqlConnectionString">the SQL connection string</param>
        static public katum ImportDatabase(IAtum atum, string rootName, DatabaseScheme dbs, string sqlConnectionString)
        {
            var root = datum.setup(atum, rootName);
            using (var connection = new SqlConnection(sqlConnectionString))
            {
                connection.Open(); // Please make sure you have a valid connection string!
                dbs.DatabaseToDatum(root, connection);

                foreach (var tableName in dbs.Keys)
                {
                    var queryString = "select * from dbo.[" + tableName + "]";
                    var command = new SqlCommand(queryString, connection);
                    IDataReader reader = command.ExecuteReader();
                    ImportTable(dbs[tableName], reader);                   
                    reader.Close();
                }
            }
            return root;
        }

        /// <summary>
        /// Import a table from an SQL database. 
        /// </summary>
        /// <param name="td">table description</param>
        /// <param name="reader">the data reader</param>
        static private void ImportTable(TableDescription td, IDataReader reader)
        {
            var tableKatum = td.TableKatum;
            var indexOfKey = td.NameOfKey == null ? -1 : reader.GetOrdinal(td.NameOfKey);

            // Import columns into columnKatums
            var columnKatums = new List<katum>();

            for (var i = 0; i < reader.FieldCount; i++)
                columnKatums.Add(td.ColumnKatums[reader.GetName(i)]);

            var ikey = 0;
            while (reader.Read())
                ikey = ImportRow(reader, tableKatum, columnKatums, indexOfKey, ikey);
        }

        /// <summary>
        /// Import a row from a table
        /// </summary>
        /// <param name="reader">data reader</param>
        /// <param name="tableKatum">the katum corresponding to the table</param>
        /// <param name="columKatums">the list of katums representing columns in the table</param>
        /// <param name="indexOfKey">the index of the key column</param>
        /// <param name="ikey">last unique key that was created, if there is no native key</param>
        /// <returns>if a key is generated, returns the incremented key</returns>
        private static int ImportRow( IDataReader reader, katum tableKatum, IReadOnlyList<katum> columKatums, int indexOfKey, int ikey)
        {
            // Each row is an instance of the tableKatum
            katum instance;
            if (indexOfKey == -1) instance = tableKatum | ikey++;             // generate a key integer
            else                  instance = tableKatum | reader[indexOfKey]; // get the key object

            for (var i = 0; i < reader.FieldCount; i++)
            {
                var fld = reader[i];
                if (fld != null && fld != DBNull.Value && indexOfKey != i)
                    instance.@is(GetAttribute(columKatums[i], fld), check:false); // no need to check graph -- to speedup
            }
            return ikey;
        }

        private static katum GetAttribute(katum property, object field)
        {
            var type = field.GetType();

            if (type == typeof(int))        return property | (int)field;
            if (type == typeof(string))     return property | (string)field;
            if (type == typeof(DateTime))   return property | (DateTime)field;
            if (type == typeof(decimal))    return property | (decimal)field;
            if (type == typeof(float))      return property | (float)field;
            if (type == typeof (short))     return property | (short) field;
            if (type == typeof(bool))       return property | (bool)field;

            if (type == typeof(double))     return property | (double)field;
            if (type == typeof(char))       return property | (char)field;
            if (type == typeof(byte))       return property | (byte)field;
            if (type == typeof(sbyte))      return property | (sbyte)field;
            if (type == typeof (ushort))    return property | (ushort)field;
            if (type == typeof(uint))       return property | (uint)field;
            if (type == typeof(long))       return property | (long)field;
            if (type == typeof(ulong))      return property | (ulong)field;

            // as object
            return property | field;
        }
    }

    /// <summary>
    /// A holder of the SQL Database Table Descriptions and the corresponding Datum Universe katums
    /// </summary>
    public class DatabaseScheme : Dictionary<string, TableDescription>
    {
        /// <summary>
        /// Constructs the dictionary from an array of table descriptions
        /// </summary>
        /// <param name="tableDescriptions"></param>
        public DatabaseScheme(TableDescription[] tableDescriptions)
        {
            foreach(var td in tableDescriptions)
                Add(td.TableName, td);
        }

        /// <summary>
        /// Create a datum graph corresponding to the table description in the dictionary.
        /// - Adds a Table katum and a Column katum under the root.
        /// - Creates a katum for each table under the Table katum
        /// - Creates a katum for each table under the Column katum 
        /// - Adds a katum for each column under the Table Column katum
        /// - Takes into consideration Keys, and Foreign Keys. 
        /// </summary>
        /// <param name="root">database root katum</param>
        /// <param name="connection">the SqlConnection to extract column and key info from the database</param>
        public void DatabaseToDatum(katum root, SqlConnection connection)
        {
            var tables = DbGetTables(connection);
            katum tableRoot = root | "Table";
            katum columnRoot = root | "Column";

            foreach (var tableName in tables)
            {
                if (!ContainsKey(tableName)) continue;

                var td = this[tableName];
                // create a katum for each table
                td.TableKatum = tableRoot | td.MappedName(tableName);
                
                // for each table create a katum for each column
                var columns = DbGetColumns(connection, tableName);

                var tableColumns = columnRoot | td.MappedName(tableName);
                td.TableKatum.@is(tableColumns);

                td.NameOfKey = null;
                foreach (var columnName in columns)
                {
                    katum columnK;
                    string fTable;

                    if (td.IsSingleKey(columnName))
                    {
                        columnK = td.TableKatum;
                        td.NameOfKey = columnName;
                    }
                    else if (td.IsFKey(columnName, out fTable))
                    {
                        if (fTable == tableName)
                        { 
                            columnK = tableColumns | td.MappedName(columnName);
                            columnK.@is(this[fTable].TableKatum);
                        }
                        else
                            columnK = this[fTable].TableKatum;
                    }
                    else
                    {
                        columnK = tableColumns | td.MappedName(columnName);
                    }

                    td.ColumnKatums.Add(columnName, columnK);
                }
            }
        }

        /// <summary>
        /// Get the katum representing a table, given the table name
        /// </summary>
        /// <param name="tableName"></param>
        /// <returns></returns>
        public string TableKatumName(string tableName)
        {
            var entityName = tableName;
            if (this[tableName].NameMap.ContainsKey(entityName))
                entityName = this[tableName].NameMap[entityName];
            
            return entityName;
        }

        /// <summary>
        /// Get a list of table names from an SqlConnection
        /// </summary>
        /// <param name="connection">the SqlConnection</param>
        /// <param name="type">table type</param>
        /// <returns></returns>
        static List<string> DbGetTables(DbConnection connection, string type = "BASE TABLE")
        {
            var restrictions = new[] { null, null, null, type };
            return connection.GetSchema("Tables", restrictions).AsEnumerable().Select(s => s.Field<string>("TABLE_NAME")).ToList();
        }

        /// <summary>
        /// Get a list of column names of a specific table from an SqlConnection
        /// </summary>
        /// <param name="connection">the SqlConnection</param>
        /// <param name="tableName">the table name</param>
        /// <returns>A list of tuples of column name and column type</returns>
        static IEnumerable<string> DbGetColumns(DbConnection connection, string tableName)
        {
            var restrictions = new[] { null, null, tableName, null };
            return connection.GetSchema("Columns", restrictions).AsEnumerable().Select(colRow => colRow.Field<string>("Column_Name")).ToList();
        }
    }

    /// <summary>
    /// Description of a table and the corresponding katum objects
    /// </summary>
    public class TableDescription
    {
        // Table Description
        public string TableName;
        public string[] KeyColumns;
        public ForeignKey[] ForeignKeys;
        public Dictionary<string, string> NameMap = new Dictionary<string, string>();
 
        // Generated katum elements
        public katum TableKatum;
        public Dictionary<string,katum> ColumnKatums = new Dictionary<string,katum>();
        public Dictionary<string,Type> ColumnTypes = new Dictionary<string, Type>(); 
        public string NameOfKey = null;

        /// <summary>
        /// Constructs a TableDescription object
        /// </summary>
        /// <param name="tableName">table name</param>
        /// <param name="tableKatumName">the name given to the corresponding katum</param>
        /// <param name="keyColumns">list of key columns, usually one column</param>
        /// <param name="foreignKeys">list of Foreign Key columns in the table</param>
        public TableDescription(string tableName, string tableKatumName, string[] keyColumns, ForeignKey[] foreignKeys = null )
        {
            TableName = tableName;
            KeyColumns = keyColumns;
            ForeignKeys = foreignKeys;

            NameMap.Add(tableName, tableKatumName);
        }

        /// <summary>
        /// Get the mapped name for this name  
        /// </summary>
        /// <param name="name">the original name</param>
        /// <returns>mapped name if exists or the original name</returns>
        public string MappedName(string name)
        {
            return NameMap.ContainsKey(name) ? NameMap[name] : name;
        }

        /// <summary>
        /// Checks if this column is the key or part of the key
        /// </summary>
        /// <param name="name">column name</param>
        /// <returns>true if the given column name is the key or part of the key</returns>
        public bool IsKey(string name)
        {
            return KeyColumns.Contains(name);
        }

        /// <summary>
        /// Checks if the given column name is the single Key
        /// </summary>
        /// <param name="name">column name</param>
        /// <returns>true if given column name is the key</returns>
        public bool IsSingleKey(string name)
        {
            return KeyColumns.Length == 1 && IsKey(name);
        }

        /// <summary>
        /// Checks if the given name is a Foreign key and if so, gives the Foreign table name 
        /// </summary>
        /// <param name="name">column name</param>
        /// <param name="fTable">out parameter; the table where the given column is the primary key</param>
        /// <returns>true if the given column name is a Foreign key</returns>
        public bool IsFKey(string name, out string fTable)
        {
            if (ForeignKeys != null)
                foreach(var fkey in ForeignKeys)
                    if (fkey.ColumnName == name)
                    {
                        fTable = fkey.ForeignTable;
                        return true;
                    }

            fTable = "";
            return false;
        }      
    }

    /// <summary>
    /// a class representing a foreign key
    /// </summary>
    public class ForeignKey
    {
        public string ColumnName;
        public string ForeignTable;

        public ForeignKey(string columnName, string foreignTable)
        {
            ColumnName = columnName;
            ForeignTable = foreignTable;
        }
    }
}
